/* import-stdb	- Imports database from other services
 *
 * PTlink Services is (C) CopyRight PTlink Coders Team 1999-2000
 * http://www.ptlink.net/Coders/ - coders@PTlink.net
 * This program is distributed under GNU Public License
 * Please read the file COPYING for copyright information.
 *
 */

#include "services.h"
#include "datafiles.h"
#include "newsserv.h"
#define NickDBName "nick.db"
#define ChanDBName "chan.db"
#define DFV_NICKSERV 6	/* NickServ datafile version */
#define DFV_CHANSERV 7	/* ChanServ datafile version */

static NickInfo *nicklists[256];	/* One for each initial character */
static ChannelInfo *chanlists[256];

/* Insert a nick alphabetically into the database. */

static void alpha_insert_nick(NickInfo *ni)
{
    NickInfo *ptr, *prev;
    char *nick = ni->nick;

    for (prev = NULL, ptr = nicklists[tolower(*nick)];
			ptr && stricmp(ptr->nick, nick) < 0;
			prev = ptr, ptr = ptr->next)
	;
    ni->prev = prev;
    ni->next = ptr;
    if (!prev)
	nicklists[tolower(*nick)] = ni;
    else
	prev->next = ni;
    if (ptr)
	ptr->prev = ni;
}

/*************************************************************************/

#define SAFE(x) do {						\
    if ((x) < 0) {						\
	fprintf(stderr, "Write error on " SERVICES_DIR "/");	\
	perror(NickDBName);					\
	exit(1);						\
    }								\
} while (0)

void save_ns_dbase(void)
{
    dbFILE *f;
    int i, j;
    NickInfo *ni;
    char **autojoin;
    Memo *memos;

    if (!(f = open_db(s_NickServ, NickDBName, "w")))
	return;
    SAFE(write_int16(DFV_NICKSERV, f));
    SAFE(write_int32(DayStats.ns_total, f));
    for (i = 0; i < 256; i++) {
	for (ni = nicklists[i]; ni; ni = ni->next) {
	    SAFE(write_int8(1, f));
	    SAFE(write_buffer(ni->nick, f));
	    SAFE(write_buffer(ni->pass, f));
	    SAFE(write_string(ni->url, f));
	    SAFE(write_string(ni->email, f));
	    SAFE(write_string(ni->icq_number, f));
	    SAFE(write_string(ni->location, f));
	    SAFE(write_string(ni->last_usermask, f));
	    SAFE(write_string(ni->last_realname, f));
	    SAFE(write_string(ni->last_quit, f));
	    SAFE(write_int32(ni->time_registered, f));
	    SAFE(write_int32(ni->last_identify, f));
	    SAFE(write_int32(ni->last_seen, f));
	    SAFE(write_int32(ni->birth_date, f));
	    SAFE(write_int16(ni->status, f));
	    SAFE(write_int16(ni->crypt_method, f));	    	    	    
	    SAFE(write_int32(ni->news_mask, f));
	    SAFE(write_int16(ni->news_status, f));	    
	    if (ni->link) {
		SAFE(write_string(ni->link->nick, f));
		SAFE(write_int16(ni->linkcount, f));
		SAFE(write_int16(ni->channelcount, f));
	    } else {
		SAFE(write_string(NULL, f));
		SAFE(write_int16(ni->linkcount, f));
		SAFE(write_int32(ni->flags, f));
		SAFE(write_int32(ni->online, f));	    		
		if (ni->flags & (NI_SUSPENDED|NI_OSUSPENDED))
		    SAFE(write_int32(ni->suspension_expire, f));
		SAFE(write_int16(ni->ajoincount, f));
		for (j=0, autojoin=ni->autojoin; j<ni->ajoincount; j++, autojoin++)		    
		    SAFE(write_string(*autojoin, f));    
		SAFE(write_int16(ni->memos.memocount, f));
		SAFE(write_int16(ni->memos.memomax, f));
		memos = ni->memos.memos;
		for (j = 0; j < ni->memos.memocount; j++, memos++) {
		    SAFE(write_int32(memos->number, f));
		    SAFE(write_int16(memos->flags, f));
		    SAFE(write_int32(memos->time, f));
		    SAFE(write_buffer(memos->sender, f));
		    SAFE(write_string(memos->text, f));
		}
		SAFE(write_int16(ni->notes.count, f));
		SAFE(write_int16(ni->notes.max, f));		
		for (j = 0; j < ni->notes.count; j++) {
		    SAFE(write_string(ni->notes.note[j], f));
		}
		SAFE(write_int16(ni->channelcount, f));
		SAFE(write_int16(ni->channelmax, f));
		SAFE(write_int16(ni->language, f));
	    }
	} /* for (ni) */
	SAFE(write_int8(0, f));
    } /* for (i) */
    close_db(f);
}
#undef SAFE

void load_nicks(FILE *f) {
    NickInfo *ni;
    char *tmp;
    char line[512];    
    int count=0;
    fgets(line,sizeof(line),f); /* read header */
    while (!feof(f)) {
	line[0]='\0';
	fgets(line,sizeof(line),f);
	if(strlen(line)<3)
	    break;
	ni = malloc(sizeof(NickInfo));
	ni->forcednicks = 0;
	DayStats.ns_total++;
	tmp = strtok(line," ");
	strcpy(ni->nick, tmp);
	strcpy(ni->pass, strtok(NULL," "));
	ni->url = ni->email = ni->icq_number = ni->location = NULL;
	ni->last_usermask = ni->last_quit = NULL;
	ni->last_realname = strdup("unknow on .db import");
	ni->online = 0;
	tmp = strtok(NULL," ");
	ni->time_registered = atol(tmp);
	tmp = strtok(NULL,"");
	ni->last_identify = atol(tmp);
	if(stricmp(ni->pass,"*"))
	    ni->status = 0;	    
	else
	    ni->status = NS_VERBOTEN;	    
	ni->last_seen = ni->last_identify;
	ni->news_mask = NM_ALL;
	ni->news_status = NW_WELCOME;
	ni->link = NULL;
	ni->linkcount = 0;
	ni->ajoincount = 0;
	ni->autojoin = NULL;		    
	ni->memos.memocount = 0;
	ni->memos.memomax = MSMaxMemos;
	ni->memos.memos = NULL;
	ni->channelmax = CSMaxReg;
	ni->language = DefLanguage;	
        ni->notes.max = NSMaxNotes;
        ni->notes.count = 0;
        ni->notes.note = malloc(ni->notes.max * sizeof(char*));  		    	
	ni->channelcount = 0;
	++count;
	alpha_insert_nick(ni); /* add nick to list */
    }
    printf("%i Done.\n",count);
}

NickInfo *findnick(const char *nick)
{
    NickInfo *ni;

    for (ni = nicklists[tolower(*nick)]; ni; ni = ni->next) {
	if (stricmp(ni->nick, nick) == 0)
	    return ni;
    }
    return NULL;
}

/*************************************************************************/

#define SAFE(x) do {						\
    if ((x) < 0) {						\
	fprintf(stderr, "Write error on " SERVICES_DIR "/");	\
	perror(ChanDBName);					\
	exit(1);						\
    }								\
} while (0)

void save_cs_dbase(void)
{
    dbFILE *f;
    int i, j;
    ChannelInfo *ci;
    Memo *memos;

    if (!(f = open_db(s_ChanServ, ChanDBName, "w")))
	return;
    write_int16(DFV_CHANSERV, f);
    for (i = 0; i < 256; i++) {
	int16 tmp16;

	for (ci = chanlists[i]; ci; ci = ci->next) {
	    SAFE(write_int8(1, f));
	    SAFE(write_buffer(ci->name, f));
	    if (ci->founder)
		SAFE(write_string(ci->founder->nick, f));
	    else
		SAFE(write_string(NULL, f));
	    if (ci->successor)
		SAFE(write_string(ci->successor->nick, f));
	    else
		SAFE(write_string(NULL, f));
	    tmp16 = ci->maxusers;
	    SAFE(write_int16(tmp16, f));
	    SAFE(write_int32(ci->maxtime, f));	   
	    SAFE(write_buffer(ci->founderpass, f));
	    SAFE(write_string(ci->desc, f));
	    SAFE(write_string(ci->url, f));
	    SAFE(write_string(ci->email, f));
	    SAFE(write_int32(ci->time_registered, f));
	    SAFE(write_int32(ci->last_used, f));
	    SAFE(write_string(ci->last_topic, f));
	    SAFE(write_buffer(ci->last_topic_setter, f));
	    SAFE(write_int32(ci->last_topic_time, f));
	    SAFE(write_int32(ci->flags, f));
	    tmp16 = ci->crypt_method;
	    SAFE(write_int16(tmp16, f));
	    if (ci->flags & CI_DROPPED)    {
		SAFE(write_int32(ci->drop_time, f));	    
	    }
	    tmp16 = CA_SIZE;
	    SAFE(write_int16(tmp16, f));
	    for (j = 0; j < CA_SIZE; j++)
		SAFE(write_int16(ci->levels[j], f));

	    SAFE(write_int16(ci->accesscount, f));
	    for (j = 0; j < ci->accesscount; j++) {
		SAFE(write_int16(ci->access[j].in_use, f));
		if (ci->access[j].in_use) {
		    SAFE(write_int16(ci->access[j].level, f));
		    SAFE(write_string(ci->access[j].ni->nick, f));
		    if(ci->access[j].who) 
			SAFE(write_string(ci->access[j].who->nick, f)); 
		    else
			SAFE(write_string("*", f));
		}		
	    }

	    SAFE(write_int16(ci->akickcount, f));
	    for (j = 0; j < ci->akickcount; j++) {
		SAFE(write_int16(ci->akick[j].in_use, f));
		if (ci->akick[j].in_use) {
		    SAFE(write_string(ci->akick[j].mask, f));
		    SAFE(write_string(ci->akick[j].reason, f));
		    if(ci->akick[j].who) 
			SAFE(write_string(ci->akick[j].who->nick, f));
		    else
			SAFE(write_string("*", f));
		    SAFE(write_int32(ci->akick[j].last_kick, f));
		}
	    }

	    SAFE(write_int16(ci->mlock_on, f));
	    SAFE(write_int16(ci->mlock_off, f));
	    SAFE(write_int32(ci->mlock_limit, f));
	    SAFE(write_string(ci->mlock_key, f));

	    SAFE(write_int16(ci->memos.memocount, f));
	    SAFE(write_int16(ci->memos.memomax, f));
	    memos = ci->memos.memos;
	    for (j = 0; j < ci->memos.memocount; j++, memos++) {
		SAFE(write_int32(memos->number, f));
		SAFE(write_int16(memos->flags, f));
		SAFE(write_int32(memos->time, f));
		SAFE(write_buffer(memos->sender, f));
		SAFE(write_string(memos->text, f));
	    }

	    SAFE(write_string(ci->entry_message, f));

	} /* for (chanlists[i]) */

	SAFE(write_int8(0, f));

    } /* for (i) */

    close_db(f);
}
#undef SAFE


static int def_levels[][2] = {
    { CA_AUTOOP,             5 },
    { CA_AUTOVOICE,          3 },
    { CA_AUTODEOP,          -1 },
    { CA_NOJOIN,            -1 },
    { CA_INVITE,             5 },
    { CA_AKICK,             10 },
    { CA_SET,   ACCESS_INVALID },
    { CA_CLEAR, ACCESS_INVALID },
    { CA_UNBAN,              5 },
    { CA_OPDEOP,             5 },
    { CA_ACCESS_LIST,        0 },
    { CA_ACCESS_CHANGE,      1 },
    { CA_MEMOREAD,           5 },
    { CA_PROTECT, ACCESS_INVALID },
    { CA_KICK,		    10 },            
    { CA_MEMOSEND,	    10 },
    { CA_MEMODEL,	    10 },        
    { -1 }
};

/* Insert a channel alphabetically into the database. */

static void alpha_insert_chan(ChannelInfo *ci)
{
    ChannelInfo *ptr, *prev;
    char *chan = ci->name;

    for (prev = NULL, ptr = chanlists[tolower(chan[1])];
			ptr != NULL && stricmp(ptr->name, chan) < 0;
			prev = ptr, ptr = ptr->next)
	;
    ci->prev = prev;
    ci->next = ptr;
    if (!prev)
	chanlists[tolower(chan[1])] = ci;
    else
	prev->next = ci;
    if (ptr)
	ptr->prev = ci;
}

static void reset_levels(ChannelInfo *ci)
{
    int i;

    if (ci->levels)
	free(ci->levels);
    ci->levels = malloc(CA_SIZE * sizeof(*ci->levels));
    for (i = 0; def_levels[i][0] >= 0; i++)
	ci->levels[def_levels[i][0]] = def_levels[i][1];
}

void load_chans(FILE *f) {
    ChannelInfo* ci;
    NickInfo* ni;
    char line[512];    
    int count=0;
    int newflags=0;
    int delchan;
    char *founder, *name, *founderpass, *ch;
    fgets(line,sizeof(line),f); /* read header */
    while (!feof(f)) {
	line[0]='\0';
	fgets(line,sizeof(line),f);
	while((ch=strchr(line,'\n'))) 
	    *ch='\0';
	name = strtok(line," ");
	if (!name) {
	    break;
	}
	founder = strtok(NULL," ");
	delchan=0;
	ci = malloc(sizeof(ChannelInfo));
	if(founder[0]=='*') {
	    newflags = CI_VERBOTEN;
	    ci->founder=NULL;
	} else {
	    ni = findnick(founder);
	    if(!ni) {
		printf("Deleting founderless channel %s, founder was %s\n",name,founder);
		delchan=1;
	    } else {
		if (ni->channelcount+1 > ni->channelcount)  /* Avoid wraparound */
		    ni->channelcount++;
	    }
	    ci->founder=ni;
	    newflags = 0;
	}
	ci->successor=NULL;
	strcpy(ci->name,name);				
	ci->flags = newflags;
	founderpass = strtok(NULL," ");
	ci->time_registered = atol(strtok(NULL," "));
	ci->last_used = atol(strtok(NULL," "));
	strcpy(ci->founderpass,founderpass);
	ci->levels = NULL;
	reset_levels(ci);
	ci->email = ci-> url = ci->last_topic = ci->entry_message = NULL;
	ci->maxusers = 0;
	ci->maxtime = time(NULL);
	ci->desc = strdup("please use /chanserv set #chan desc descrtiption");
	strcpy(ci->last_topic_setter,"");
	ci->last_topic_time = time(NULL);
	ci->crypt_method = 0;
	fgets(line,sizeof(line),f); /* read ACCESS tag */
	while((ch=strchr(line,'\n'))) 
	    *ch='\0';
	if(!stricmp(line,"ACCESS")) {
	    char ac_nick[1024][NICKMAX];
	    int	ac_level[1024];
	    int aci = 0;	    
	    do {
		line[0]='\0';
		fgets(line,sizeof(line),f); /* read nick level pair */
		while((ch=strchr(line,'\n'))) 
		    *ch='\0';		
		if(line[0] == '-') /* end of list */
		    break;
		strcpy(ac_nick[aci],strtok(line," ")); 
		ac_level[aci] = atoi(strtok(NULL,""));
		++aci;	/* next access*/
	    } while(line[0] != '-'); 
	    ci->accesscount = aci;
	    ci->access = aci ? malloc(ci->accesscount *sizeof(ChanAccess)): NULL;
	    
	    for (aci = 0; aci < ci->accesscount; aci++) {
		ci->access[aci].level=ac_level[aci];
		ci->access[aci].ni=findnick(ac_nick[aci]);
		ci->access[aci].in_use =  ci->access[aci].ni ? 1 : 0;
	    }
	    ++count;
	}
	ci->akickcount = 0;
	ci->akick = NULL;
	ci->mlock_on = CMODE_N | CMODE_T;
	ci->mlock_off = 0;
	ci->mlock_limit = 0;
	ci->mlock_key = 0;
	ci->memos.memocount = 0;
	ci->memos.memomax = 
	ci->memos.memomax = MSMaxMemos;
	if(!delchan)
	    alpha_insert_chan(ci);
    }
    printf("%i Done.\n",count);    
}

int main(int argc, char **argv) {
    char *datadir;
    FILE *stdb;
    char fn[255]; 
    printf("import-stdb v1.0 - Read .stdb data files and creates PTlink.Services .db\n");
    if(argc>1) {
	if(!strcasecmp(argv[1],"--version")) {
	    return 0;
	} else if(argv[1][0]=='-') {
	    printf("usage: import-stdb [stdb_directory]\n\n");
	    printf("	stdb_directory specifies nick.stdb and chan.stdb location\n");
	    printf("	current directory will be used as default\n");
	    return 0;
	}
    }
    if (!read_config(0))
	return 1;
    datadir = (argc>1) ? argv[1]: ".";
    sprintf(fn,"%s/nick.stdb",datadir);
    stdb=fopen(fn,"r");
    if(!stdb) {
	printf("Error: Could not open %s\n",fn);
	return 2;
    }
    printf("Reading nicknames info...");
    load_nicks(stdb);
    fclose(stdb);
    printf("Saving nick.db...");
    save_ns_dbase();
    printf("Done\n");
    sprintf(fn,"%s/chan.stdb",datadir);
    stdb=fopen(fn,"r");
    if(!stdb) {
	printf("Error: Could not open %s\n",fn);
	return 2;
    }
    printf("Reading channels info...");
    load_chans(stdb);
    fclose(stdb);
    printf("Saving chan.db...");
    save_cs_dbase();
    printf("Done\n");
    return 0;
}
